/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:   
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and 
 * international Copyright laws.  Users and possessors of this source code 
 * are hereby granted a nonexclusive, royalty-free license to use this code 
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE 
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR 
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH 
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS 
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE 
 * OR PERFORMANCE OF THIS SOURCE CODE.  
 *
 * U.S. Government End Users.   This source code is a "commercial item" as 
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of 
 * "commercial computer  software"  and "commercial computer software 
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995) 
 * and is provided to the U.S. Government only as a commercial end item.  
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through 
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the 
 * source code with only those rights set forth herein. 
 *
 * Any use of this source code in individual and commercial software must 
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */



//Total number of possible data values
#define      BIN_COUNT 64
#define HISTOGRAM_SIZE (BIN_COUNT * sizeof(unsigned int))

//Threads per block count, preferred to be a multiple of 64 (see below)
#define     THREAD_N 192
//Per-block number of elements in all per-thread histograms
#define BLOCK_MEMORY (THREAD_N * BIN_COUNT)
//Dwords processed by single block count
//Since counters are one-byte, in order to avoid overflow(wrap around)
//this size must be limited by the number
#define   BLOCK_DATA (THREAD_N * 63)



#define IMUL(a, b) __mul24(a, b)



////////////////////////////////////////////////////////////////////////////////
// If threadPos == threadIdx.x, there are always  4-way bank conflicts,
// since each group of 16 threads (half-warp) accesses different bytes,
// but only within 4 shared memory banks. Having shuffled bits of threadIdx.x 
// as in histogram64GPU(), each half-warp accesses different shared memory banks
// avoiding any bank conflicts at all.
// See supplied whitepaper for detailed explanations.
////////////////////////////////////////////////////////////////////////////////
__device__ void addData64(unsigned char *s_Hist, int threadPos, unsigned int data){
    s_Hist[threadPos + IMUL(data, THREAD_N)]++;
}

////////////////////////////////////////////////////////////////////////////////
// Main computation pass: compute gridDim.x partial histograms
////////////////////////////////////////////////////////////////////////////////
__global__ void histogram64Kernel(unsigned int *d_Result, unsigned int *d_Data, int dataN){
    //Global base index in input data for current block
    const int  baseIndex = IMUL(BLOCK_DATA, blockIdx.x);
    //Current block size, clamp by array border
    const int   dataSize = min(dataN - baseIndex, BLOCK_DATA);
    //Encode thread index in order to avoid bank conflicts in s_Hist[] access:
    //each half-warp accesses consecutive shared memory banks
    //and the same bytes within the banks
    const int threadPos = 
        //[31 : 6] <== [31 : 6]
        ((threadIdx.x & (~63)) >> 0) |
        //[5  : 2] <== [3  : 0]
        ((threadIdx.x &    15) << 2) |
        //[1  : 0] <== [5  : 4]
        ((threadIdx.x &    48) >> 4);

    //Per-thread histogram storage
    __shared__ unsigned char s_Hist[BLOCK_MEMORY];

    //Don't forget to clear shared memory before utilization
    //No need in per-byte access at this stage
    for(int pos = threadIdx.x; pos < (BLOCK_MEMORY / 4); pos += blockDim.x)
        ((unsigned int *)s_Hist)[pos] = 0;

    __syncthreads();
    ////////////////////////////////////////////////////////////////////////////
    // Cycle through current block, update per-thread histograms
    // Since only 64-bit histogram of 8-bit input data array is calculated,
    // only highest 6 bits of each 8-bit data element are extracted,
    // leaving out 2 lower bits.
    ////////////////////////////////////////////////////////////////////////////
    for(int pos = threadIdx.x; pos < dataSize; pos += blockDim.x){
        unsigned int data4 = d_Data[baseIndex + pos];
        addData64(s_Hist, threadPos, (data4 >>  2) & 0x3FU);
        addData64(s_Hist, threadPos, (data4 >> 10) & 0x3FU);
        addData64(s_Hist, threadPos, (data4 >> 18) & 0x3FU);
        addData64(s_Hist, threadPos, (data4 >> 26) & 0x3FU);
    }

    __syncthreads();
    ////////////////////////////////////////////////////////////////////////////
    // Merge per-thread histograms into per-block and write to global memory.
    // Start accumulation positions for half-warp each thread are shifted
    // in order to avoid bank conflicts. 
    // See supplied whitepaper for detailed explanations.
    ////////////////////////////////////////////////////////////////////////////
    if(threadIdx.x < BIN_COUNT){
        unsigned int sum = 0;
        const int value = threadIdx.x;

        const int valueBase = IMUL(value, THREAD_N);
        const int  startPos = IMUL(threadIdx.x & 15, 4);

        //Threads with non-zero start positions wrap around the THREAD_N border
        for(int i = 0, accumPos = startPos; i < THREAD_N; i++){
            sum += s_Hist[valueBase + accumPos];
            if(++accumPos == THREAD_N) accumPos = 0;
        }

#if ATOMICS
        atomicAdd(d_Result + value, sum);
#else
        d_Result[IMUL(BIN_COUNT, blockIdx.x) + value] = sum;
#endif
    }
}



////////////////////////////////////////////////////////////////////////////////
// Merge blockN histograms into gridDim.x histograms
// blockDim.x == BIN_COUNT
// gridDim.x  == BLOCK_N2
////////////////////////////////////////////////////////////////////////////////
__global__ void mergeHistogram64Kernel(unsigned int *d_Result, int blockN){
    const int  globalTid = IMUL(blockIdx.x, BIN_COUNT) + threadIdx.x;
    const int numThreads = IMUL(gridDim.x,  BIN_COUNT);
    const int   dataSize = IMUL(blockN,     BIN_COUNT);

    int sum = 0;
    for(int pos = globalTid; pos < dataSize; pos += numThreads)
       sum += d_Result[pos];

    d_Result[globalTid] = sum;
}



////////////////////////////////////////////////////////////////////////////////
// CPU interface to GPU histogram calculator
////////////////////////////////////////////////////////////////////////////////
//histogram64kernel() results buffer
unsigned int *d_Result64;
unsigned int *h_Result64GPU;

//Maximum block count for histogram64kernel()
//Limits input data size to 756MB
const int MAX_BLOCK_N = 16384;

//Internal memory allocation
const int BLOCK_N2 = 32;

void initHistogram64(void){
    h_Result64GPU = (unsigned int *)malloc(HISTOGRAM_SIZE * BLOCK_N2);
#ifdef ATOMICS
    CUDA_SAFE_CALL( cudaMalloc((void **)&d_Result64, HISTOGRAM_SIZE) );
#else
    CUDA_SAFE_CALL( cudaMalloc((void **)&d_Result64, MAX_BLOCK_N * HISTOGRAM_SIZE) );
#endif
}

//Internal memory deallocation
void closeHistogram64(void){
    CUDA_SAFE_CALL( cudaFree(d_Result64) );
}

void histogram64GPU(
    unsigned int *h_Result,
    unsigned int *d_Data,
    int dataN
){
    const int  blockN = iDivUp(dataN / 4, THREAD_N * 63);
    if(blockN > MAX_BLOCK_N){
        printf("histogram64gpu(): data size exceeds maximum\n");
        return;
    }

    #ifdef ATOMICS
        CUDA_SAFE_CALL( cudaMemset(d_Result64, 0, HISTOGRAM_SIZE) );
        histogram64Kernel<<<blockN, THREAD_N>>>(
            d_Result,
            (unsigned int *)d_Data,
            dataN / 4
        );
        CUT_CHECK_ERROR("histogram64Kernel() execution failed\n");
        CUDA_SAFE_CALL( cudaMemcpy(h_Result, d_Result64, HISTOGRAM_SIZE, cudaMemcpyDeviceToHost) );
    #else
        histogram64Kernel<<<blockN, THREAD_N>>>(
            d_Result64,
            (unsigned int *)d_Data,
            dataN / 4
        );
        CUT_CHECK_ERROR("histogram64Kernel() execution failed\n");

        const int FINAL_BLOCK_N = min(BLOCK_N2, blockN);
        if(BLOCK_N2 < blockN){
            mergeHistogram64Kernel<<<BLOCK_N2, BIN_COUNT>>>(
                d_Result64,
                blockN
            );
            CUT_CHECK_ERROR("mergeHistogram64Kernel() execution failed\n");
        }

        CUDA_SAFE_CALL( cudaMemcpy(h_Result64GPU, d_Result64, FINAL_BLOCK_N * HISTOGRAM_SIZE, cudaMemcpyDeviceToHost) );
        for(int i = 1; i < FINAL_BLOCK_N; i++){
            for(int j = 0; j < BIN_COUNT; j++)
                h_Result64GPU[j] += h_Result64GPU[i * BIN_COUNT + j];
        }
        for(int i = 0; i < BIN_COUNT; i++)
            h_Result[i] = h_Result64GPU[i];

    #endif
}
